#include "graph_lib.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>

#define LOG_DEBUG 1
#define LOG_WARN 2
#define LOG_ERR 3
#define LOG_LEVEL LOG_WARN


//get only filename (without path) from argument for input file
char* get_only_filename( char* filename_with_path ) {	
	int i=0;
	while (filename_with_path[i]!='\0') i++;
	while (filename_with_path[i]!='/' && i>0) i--;
	if (i>0) i++;
	return &filename_with_path[i];
}


int main(int argc, char *argv[])
{
	int max_degree, num_colors; 
	
	if (argc != 3 && argc != 4) {
		fprintf( stderr, "USAGE: ./graph input_file number_of_colors [number_of_a_graph]\n" );
		fprintf( stderr, "\tit is assumed that the graph has a vertex of degree D, \n\twhere number_of_colors-1 <= D <= number_of_colors\n" );
		fprintf( stderr, "\n\tif number_of_a_graph is given, only one graph is processed\n" );
		return 0;
	} else {
		num_colors = atoi( argv[2] );
		if (num_colors<1) {
			fprintf( stderr, "       arguments must be positive integers" );
			return 0;
		}
	}

	fprintf( stderr, "opening %s\n", argv[1] );
	FILE *input_file = fopen( argv[1], "r" );
	if (NULL == input_file ) {
		msg (LOG_ERR, "failed to open input file for reading ");
		exit(1);	
	}; 
	
	
	char *filename = get_only_filename( argv[1] );
		
	
	char output_meta_file_name[60];
	sprintf( output_meta_file_name, "output/1VERTEX_col%i_%s", num_colors, filename );
	
	long graph_offset = 0;
	int is_num_graph_given;
	if( argc == 4 ) {
		is_num_graph_given = 1;
		graph_offset = atoi( argv[3] );
	} else {
		is_num_graph_given = 0;
		graph_offset = count_num_lines( output_meta_file_name );
	}
		
	skip_lines( graph_offset, input_file );

	//create empty meta file if it doesn't exist
	if ( !is_num_graph_given ) {
		FILE *output_meta_file = fopen( output_meta_file_name, "a" );
		//if meta file doesn't contain any graph results, then print header to meta file	
		if ( graph_offset == 0 ) {
			fprintf( output_meta_file, "graph# (index to %s)\tnum_vertices\tmax_degree\tnum_edges\tcolorings\tcomponents\tbipartiteness\n", filename );
		}

		fclose( output_meta_file );
	}


	graph_t *graph, *linegraph, *graph_of_colorings; 
	coloring_t *coloring;
	code_properties_t code_properties;
	unsigned long num_colorings = 0;
	char **dictionary;
	
	long count_graphs = graph_offset;
	
	fprintf( stderr, "is_num_graph_given : %d\n", is_num_graph_given );
	
	
	//loop for all file lines (i.e. all graphs
	while( 	NULL != ( graph = g_create_from_file( input_file ) ) ) { 	
		
		//if specific graph number is given, then only one graph should be
		//processed. If loop wants to process other graph, break the loop
		if ( is_num_graph_given && count_graphs != graph_offset ) {
			break;
		}
		
		//if graph doesn't contain vertex of degree max_degree, then skip it
		//else convert it, so that vertex 0 has "max_degree" neighbours
		max_degree = g_max_degree( graph );
		
		if ( num_colors < max_degree ) {
			g_destroy( &graph );
			
/*			if ( !is_num_graph_given ) {
				FILE *output_meta_file = fopen( output_meta_file_name, "a" );
				if (NULL == output_meta_file ) {
					msg (LOG_ERR, "failed to open metafile for writing ");
					exit(1);	
				}; 
		
				fprintf( output_meta_file, "%ld\tskipped\n", count_graphs );
		
				fclose( output_meta_file );
			}
*/			

			count_graphs++;
			continue;
		} else {
			
			int max_degree_index = 0;
			while ( g_vertex_degree( graph, max_degree_index) < max_degree ) {
				max_degree_index++;
			}
			
			if (max_degree_index != 0 ) {
				
				graph_t *tmp_graph = g_create( graph->num_vertices );
				int i, j;
				for ( i = 0; i<graph->num_vertices-1; i++ ) {
					for ( j = i+1; j<graph->num_vertices; j++ ) {
						if (g_is_edge( graph, i, j ) ) {
							int new_i, new_j;
							
							if (i == 0) new_i = max_degree_index;
							else if (i == max_degree_index) new_i = 0;
							else new_i = i;


							if (j == 0) new_j = max_degree_index;
							else if (j == max_degree_index) new_j = 0;
							else new_j = j;
							g_add_edge( tmp_graph, new_i, new_j );
						}
					}
				}
				
				g_destroy( &graph );
				graph = tmp_graph;
			}
		}
				
		linegraph = g_create_linegraph( graph );

		nv_t i;
		
		coloring = g_create_coloring( num_colors, linegraph->num_vertices );		
		
		code_properties = g_code_properties( num_colors, linegraph->num_vertices );
		
		//count number of colorings
		num_colorings = 0;

		//uncomment this to get all colorings, not only standard ones
		//max_degree = 0;
		
		g_init_coloring_for_regular_graph( coloring, 0 );
		
		int tmp;
		while ( ( tmp = g_get_next_coloring_offset( linegraph, coloring, 0 ) ) > 0 ) {
			num_colorings++;
			if (tmp==2) break;			
		}

		if (num_colorings == 0) {
			count_graphs++;
			g_destroy_coloring( &coloring );
			g_destroy( &linegraph );
			g_destroy( &graph );
			continue;
		}
		
		
		//create graph of colorings
		graph_of_colorings = g_create( num_colorings );
	

		//set coloring prefix
		g_init_coloring_for_regular_graph( coloring, 0 );
		
		//initialize dictionary
		dictionary = g_create_dictionary( num_colorings );
		num_colorings=0;
		//store all colorings to dictionary
		while ( (tmp = g_get_next_coloring_offset( linegraph, coloring, 0 ) ) > 0 ){
			dictionary[num_colorings] = g_create_code( &code_properties );
			g_coloring_to_code( coloring, dictionary[num_colorings], &code_properties );
			g_code_to_coloring( dictionary[num_colorings], coloring, &code_properties );
			num_colorings++;
			if (tmp == 2) break;
		}	
	
		
		//perform switching
		for (i=0; i<num_colorings; i++) {
			g_code_to_coloring( dictionary[i], coloring, &code_properties );
			perform_1vertex_switching_on_coloring( linegraph, graph_of_colorings, dictionary, coloring );
		}
		
		
		//output results
		if ( is_num_graph_given ) {
			fprintf( stderr, "num_colorings: %lu\n", num_colorings );
			g_dump_goc_graphml( graph, graph_of_colorings, linegraph, dictionary, num_colors );
		}

		if ( !is_num_graph_given ) {
			FILE *output_meta_file = fopen( output_meta_file_name, "a" );
			if (NULL == output_meta_file ) {
				msg (LOG_ERR, "failed to open metafile for writing ");
				exit(1);	
			}; 
	
			fprintf( output_meta_file, "%ld\t%d\t%d\t%d\t%lu\t%lu\t%i\n", count_graphs, graph->num_vertices, g_max_degree(graph), linegraph->num_vertices, num_colorings, g_num_components( graph_of_colorings ), g_bipartiteness( graph ) );
	
			fclose( output_meta_file );
		}
		
//		g_dump_goc_graphml( graph, graph_of_colorings, linegraph, dictionary, num_colors );
//		g_dump_wilma( graph_of_colorings );
//		exit(0);


		
		for (i=0; i<num_colorings; i++) {
			free(dictionary[i]);
		}
		g_destroy_dictionary( &dictionary );
		g_destroy_coloring( &coloring );
		g_destroy( &graph );
		g_destroy( &linegraph );
		g_destroy( &graph_of_colorings );
		count_graphs++;
	
	}
	
	fclose( input_file );
	
	
	fprintf( stderr, "alles gute\n\n\n");
	
	return 0;

}
