#include "graph_lib.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>

#define LOG_DEBUG 1
#define LOG_WARN 2
#define LOG_ERR 3
#define LOG_LEVEL LOG_WARN


int combin[10][3] = {
	{ 0, 1, 2 },  //0
	{ 0, 1, 3 },  //1
	{ 0, 1, 4 },  //2
	{ 0, 2, 3 },  //3
	{ 0, 2, 4 },  //4
	{ 0, 3, 4 },  //5
	{ 1, 2, 3 },  //6
	{ 1, 2, 4 },  //7
	{ 1, 3, 4 },  //8
	{ 2, 3, 4 },  //9
};


int complement[10][2] = {
	{ 3, 4 },
	{ 2, 4 },
	{ 2, 3 },
	{ 1, 4 },
	{ 1, 3 },
	{ 1, 2 },
	{ 0, 4 },
	{ 0, 3 },
	{ 0, 2 },
	{ 0, 1 },
};

void set_prefix( coloring_t *coloring, int num_vertices_in_prefix, int num_edges, int *indexes ) {
	
	int i, j;
	
	for ( i = 0; i < num_vertices_in_prefix; i++ ) {
		for ( j = 0; j < 3; j++ ) {
			coloring->colors[3*i+j] = combin[indexes[i]][j];
		}
	}
	
	for ( i = num_vertices_in_prefix * 3; i < num_edges; i++ ) {
		coloring->colors[i] = -1;
	}
}


int is_triangle( int *string ) {
	int counts[5];
	int i, number_of_ones = 0, number_of_twos = 0;
	for ( i = 0; i < 5; i++ ) {
		counts[i]=0;	
	}
	
	for ( i = 0; i < 4; i++ ) {
		counts[	string[i] ]++;
	}
	
	for ( i = 0; i < 5; i++ ) {
		if (counts[i] == 1) number_of_ones++;
		if (counts[i] == 2) number_of_twos++;
	}

	if (number_of_ones != 2 || number_of_twos != 1) {
		return 0;
	}
	
	for ( i = 4; i < 6; i++ ) {
		counts[	string[i] ]++;
	}

	number_of_ones = 0;
	number_of_twos = 0;
	
	for ( i = 0; i < 5; i++ ) {
		if (counts[i] == 1) number_of_ones++;
		if (counts[i] == 2) number_of_twos++;
	}
	
	if (number_of_ones != 0 || number_of_twos != 3) {
		return 0;
	}
	
	return 1;
}


int is_cycle( graph_t *mgraph ) {
		//compute number of edges
	int num_edges = 0, j, k;
	for (j = 0; j<4; j++) {
		for (k = j; k<5; k++) {
			if (g_is_edge( mgraph, j, k ) ) {
				num_edges++;
			}
		}
	}
				
	if( num_edges == 4 ) {
		int num_components = g_num_components( mgraph );
		
		if ( num_components == 2 ) {
			return 1;
		} else if( num_components == 1 ) {
			return 0;
		} else {
			msg( LOG_ERR, "impossible number of components. Something bad happened" );
			exit(1);
		}
	} else if (num_edges == 3) {
		//if 3 vertices have degree 2 then there is a triangle
		int num_deg2 = 0, num_deg0 = 0;
		
		for (j = 0; j<5; j++) {
			int current_deg = 0;
			for (k = 0; k<5; k++) {
				//no loops possible
				if (j==k) continue;
				
				if (g_is_edge( mgraph, j, k ) ) {
					current_deg++;
				}
				if (current_deg==2) num_deg2++;
				else if (current_deg==0) num_deg0++;
				else {
					//some vertex has other degree than 0 or 2, meaning there
					//is no cycle	
					return 0;
				}
			}
		}
		
		if (num_deg2 == 3 && num_deg0 == 2 ) {
			return 1;
		}
	}
	
	return 0;
}



int main(int argc, char *argv[])
{
	int num_free_edges = 6;
	int num_core_vertices = 3;

	
	
	graph_t *graph = g_create( num_free_edges*4 + num_core_vertices );

	//create graph and enumerate edges so that edges neighbouring with central
	//vertices are the last

	int i, j, num_edges = 0;
	for (i = 0; i<num_free_edges; i++ ) {
		int vertex1 = i * 4;
		for (j = 1; j<4; j++ ) {
			int vertex2 = vertex1 + j;
			g_add_edge( graph, vertex1, vertex2 );
			g_set_data_both_directions( graph, num_edges++, vertex1, vertex2 );
		}
	}

	int current_vertex = num_free_edges*4;
	int current_free_edge = 0;

	//here add the core graph
	i = current_vertex;	
	g_add_edge( graph, i+0, i+1 );   g_set_data_both_directions( graph, num_edges++, i+0, i+1  );
	g_add_edge( graph, i+1, i+2 );   g_set_data_both_directions( graph, num_edges++, i+1, i+2  );
	g_add_edge( graph, i+2, i+0 );   g_set_data_both_directions( graph, num_edges++, i+2, i+0  );
	
	
	//connect core to surrounding
	for (i=current_vertex; i < current_vertex + num_core_vertices; i++) {
		int free_edges = 4 - g_vertex_degree( graph, i );
		for( j=0; j < free_edges; j++ ) {
			g_add_edge( graph, i, current_free_edge );
			g_set_data_both_directions( graph, num_edges++, i, current_free_edge  );
			current_free_edge += 4;
		}
	}

	

	// get neighbours of one of core vertices so that I can skip some duplicites later
	int neighbours_of_core_vertex[4];
	{
		
		edge_iterator_t iterator;
		edge_t *edge;
	
		iterator = g_new_edge_iterator( graph );
		iterator.vertex = current_vertex;
	
		i = 0;
		while ( NULL != (edge = g_edge_iterator_next_for_vertex( &iterator ) ) ) {
			neighbours_of_core_vertex[i] = g_get_data( graph, current_vertex, edge->vertex ); 
			printf("%d ", neighbours_of_core_vertex[i] );
			i++;
		}
	}
	
	
	//create linegraph
	graph_t *linegraph = g_create_linegraph_from_enumerated_graph( graph, num_edges );
	
	graph_t *graph_of_colorings;	

	coloring_t *coloring = g_create_coloring( 5, linegraph->num_vertices );
	
	code_properties_t code_properties = g_code_properties( 5, linegraph->num_vertices ); //num_colors = 5

	
	int num_colorings=0, number_of_non_kempe_equivalent = 0, number_of_kempe_equivalent = 0;

	char **dictionary;
	
	int indexes[6] = {0, 0, 0, 0, 0, 0};
	
	
	
	//for every configuration
	for ( i = 0; i < 1000000; i++ ) {
		if (i%10000 == 0) fprintf( stderr, "%i\n", i );	


		//set current indexes
		int val = i;
		for (j = 5; j>=0; j-- ) {
			indexes[j] = val % 10;
			val = val/10;
		}
		
		set_prefix( coloring, num_free_edges, num_edges, indexes );

		//the coloring is correct when the colors of edges around central vertex are 0,1,2,3
		int correct_coloring = 0;

		num_colorings = 0;
		while( g_get_next_coloring_offset( linegraph, coloring, num_edges-9 ) ) {
			if ( ( coloring->colors[neighbours_of_core_vertex[0]] == 0 ) &&
				 ( coloring->colors[neighbours_of_core_vertex[1]] == 1 ) &&
				 ( coloring->colors[neighbours_of_core_vertex[2]] == 2 ) &&
				 ( coloring->colors[neighbours_of_core_vertex[3]] == 3 ) ) {
				correct_coloring = 1;
			}

			num_colorings++;
		}
		
		if ( correct_coloring == 0 ) { 
			continue;
		}
		
		set_prefix( coloring, 6, num_edges, indexes );
		
		dictionary = g_create_dictionary( num_colorings );
		graph_of_colorings = g_create( num_colorings );

		num_colorings=0;
		while ( g_get_next_coloring_offset( linegraph, coloring, num_edges-9 ) ) {
			dictionary[num_colorings] = g_create_code( &code_properties );
			g_coloring_to_code( coloring, dictionary[num_colorings], &code_properties );
			num_colorings++;
		}
	
		int m;
		//perform switching
		for (m=0; m<num_colorings; m++) {
			g_code_to_coloring( dictionary[m], coloring, &code_properties );
			perform_kempe_switching_on_coloring( linegraph, graph_of_colorings, dictionary, coloring, num_edges-7 );
		}
		
		nv_t nc = g_num_components( graph_of_colorings );

		if (nc<2) {
			printf( "%d %lu \n", i, nc );
			number_of_kempe_equivalent++;
		} else {
			number_of_non_kempe_equivalent++;
		}
		
//			g_dump_goc_graphml( graph, graph_of_colorings, linegraph, dictionary, 5 ); //num_colors =  5
//			g_dump_goc_mg_graphml( graph, graph_of_colorings, linegraph, dictionary, 5, indexes ); //num_colors =  5
//exit(0);
		
		
		//destroy dictionary and goc - avoid memory leak
		for (m=0; m<num_colorings; m++) {
			free(dictionary[m]);
		}
		g_destroy_dictionary( &dictionary );
		
		g_destroy( &graph_of_colorings );
		
		//and let's go to another coloring prefix		
	}
	
		
	printf("number_of_non_kempe_equivalent = %d\n", number_of_non_kempe_equivalent );
	printf("number_of_kempe_equivalent = %d\n", number_of_kempe_equivalent );

	g_destroy( &graph );
	g_destroy( &linegraph );
	g_destroy_coloring( &coloring );

	
	
	
	
	return 0;

}
